import React, { Component } from 'react';
import { isEmpty,
         cloneDeep,
         pickBy, 
         isEqual,
         isElement,
         isNil            } from 'lodash';

/**
 * @typedef {Object} ISelection - Represents a selection object
 * @property {number} containerIndex - The container index
 * @property {number=} childIndex - The child index
 * @property {number=} blockIndex - The block index
 */

/**
 * @typedef {Object} IFocusRectangle - Represents a focus rectangle object
 * @property {number} width
 * @property {number} height
 * @property {number} left
 * @property {number} top 
 */

/**
 * @typedef {Object} FocusRectangleProps - Represents the props of this component
 * @property {ISelection} Selection - the selection of this component
 * @property {HTMLElement} container - The container we are going to be searching through
 * @property { () => IFocusRectangle } focusCalculated - callback to execute whenever this component is done with the calculations of the focus rectangle
 */

/**
 * @typedef {'container' | 'child' | 'block'} TElementType - Represents the type of elements that are possible
 */

/**
 * This component will be responsible for showing a focus rectangle around elements
 */
class FocusRectangle extends Component {

    /** 
     * Component reference to this element 
     * @type {HTMLElement}
     */
    el = null;

    /**
     * Set height of this component
     * @param {number} val
     */
    set height(val) {
        this.el.style.height = `${val}px`;
    }

    /**
     * Set width of this component
     * @param {number} val
     */
    set width(val) {
        this.el.style.width = `${val}px`;
    }

    /**
     * Set left property of this component
     * @param {number} val
     */
    set left(val) {
        this.el.style.left = `${val}px`;
    }
    
    /**
     * Set top property of this component
     * @param {number} val
     */
    set top(val) {
        this.el.style.top = `${val}px`;
    }
    
    /**
     * State of this component
     * @type { FocusRectangleProps }
     */
    props;

    componentDidMount() {
        console.log('[focus-rectangle] component mounted', cloneDeep(this.props));

        this.updateSelection(this.props.selection);
    }

    componentWillReceiveProps(nextProps) {
        // We only want to update the selection if the props change
        //
        if (!isEqual(nextProps, this.props)) {
            this.updateSelection(nextProps.selection);    
        }
    }

    /**
     * Updates the selection properly
     * @param {ISelection} selection
     */
    updateSelection(selection) {
        // Means this method was called too early, so we will exit early
        //
        if ( !this.el || !this.props.container ) {
            return;
        }

        // Incase the selection was reset
        //
        if ( isNil(selection) ) {
            console.log('[focus-rectangle] element is ', this.el);
            this.left = this.top = this.width = this.height = 0;
        } else {
            this.selectElement(selection);
        }
    }

    /**
     * Selects an element
     * @param {ISelection} selection
     */
    selectElement(selection) {
        /** 
         * Resulting HTML element
         * @type {HTMLElement}
         */
        let result= null;

        const elementType = determineElementType(selection);

        if (elementType === 'container') {
            result = this.props.container.querySelector(`.eva-container-${selection.containerIndex}`);
        }
        if (elementType === 'child') {
            result = this.props.container.querySelector(`.eva-container-${selection.containerIndex} .eva-child-${selection.childIndex}`);
        }
        if (elementType === 'block') {
            result = this.props.container.querySelector(`.eva-container-${selection.containerIndex} .eva-child-${selection.childIndex} .eva-visual-block-${selection.blockIndex}`);
        }

        if (isElement(result)) {
            const rectangle = this.createFocusRectangle(result);

            this.top = rectangle.top;
            this.left = rectangle.left;
            this.width = rectangle.width;
            this.height = rectangle.height;
        } else {
            console.warn('[focus-rectangle] failed to find an element using the following result', result);
        }
    }

    /**
     * Generic method to paint a blue rectangle on the canvas
     * @param {HTMLElement} element
     * @author janvandenberg
     * @returns { IFocusRectangle }
     */
    createFocusRectangle(element)  {
        const elemRect = element.getBoundingClientRect();
        const elemtCompStyle = window.getComputedStyle(element);
        const offsetTop = elemRect.top;
        const offsetLeft = elemRect.left;

        const focusRectangle = {
              width: element.offsetWidth
            , height: element.offsetHeight
            , left: offsetLeft + window.scrollX
            , top: offsetTop + window.scrollY
        };

        this.props.focusCalculated(focusRectangle);

        return focusRectangle;
    }
    
    render() {
        if ( !this.props.selection ) {
            return (null);
        } else {
            return (
                <div className="FocusRectangle" ref={(container) => this.el = container} style={{ backgroundColor: '#a0c5e9', position: 'absolute', opacity: 0.6, pointerEvents: 'none' }}>
                </div>
            )
        }
    }
}

export default FocusRectangle;


/**
 * Determines what kind of element is selected based on the current selection
 * @param {ISelection} param
 * @returns {TElementType}
 */
export const determineElementType = (selection) => {
    /** Filtered out keys that are index related  */
    const keys = Object.keys(selection);

    let elementType = null;

    if (keys.length === 1) { elementType = 'container'; };

    if (keys.length === 2) { elementType = 'child'; };

    if (keys.length === 3) { elementType = 'block'; };

    return elementType;
};