//@ts-check
import React, { Component, MouseEvent } from 'react';
import { isEmpty, cloneDeep, pickBy, isEqual, isArray, get } from 'lodash';
import FocusRectangle from './focus-rectangle/focus-rectangle';
import withStyles from 'isomorphic-style-loader/lib/withStyles';
import styles from './preview.css';
import VisualBlock from '../visual-block/visual-block';
import Counter from '../counter/counter';
import Header from '../header/header';

/**
 * @typedef {Object} ISelection - Represents a selection object
 * @property {number} containerIndex - The container index
 * @property {number=} childIndex - The child index
 * @property {number=} blockIndex - The block index
 */

 /**
 * @typedef {Object} PreviewProps - Represents the properties the preview has
 * @property { { [key: string]: any } } components - A list of components the preview can render
 * @property {string} endPointUrl - The end point URL of the current environmnet
 */



class Preview extends Component {


    get components() {
        return this.props.components || {};
    }

    /** 
     * Component reference to this element 
     * @type {HTMLElement}
     */
    el = null;

    /** 
     * Reference to the focus rectangle
     * @type {FocusRectangleProps}
     */
    focusRectangle = null;


    /**
     * The default props of this component
     * @type {PreviewProps}
     */
    static defaultProps = {
        endPointUrl: null,
        components: {}
    }

    constructor() {
        super();
        
        /**
         * State of this component
         * @type { { page: any, selection: ISelection, container: HTMLElement, focusRectangleKey: number } }
         */
        this.state = {
            page: null,
            selection: null,
            container: null,
            focusRectangleKey: Math.random()
        }
    }

    componentDidMount() {
    
        window.addEventListener('resize', this.handleResize.bind(this));

        this.postToEditor('onInit', true);

        window.addEventListener('message', this.listenForMessages.bind(this) );

        window.addEventListener('keyup', this.listenForKeyUp.bind(this) )

        window.addEventListener('click', this.listenForClick.bind(this) )

        this.handleResize();
    }

    /**
     * This function will be responsible for handling key up events and will propgate that back to the editor
     */
    listenForKeyUp = (event) => {
        /** This object will contain the properties we are interested in */
        let {
            shiftKey,
            ctrlKey,
            metaKey,
            altKey,
            key
            } = event;
        
        this.postToEditor('onKeyUp', {
            shiftKey,
            ctrlKey,
            metaKey,
            altKey,
            key
        });
    }

    listenForMessages = (e) => {
        console.log('[preview] received message', e);
        if ( isArray(e.data) ) {
            e.data.forEach(data => {
                console.log('[preview] message obj ', data);
                console.log('[preview] received a message on ', data.channel);
                console.log('[preview] message is ', data.message);

                if (data.channel === 'onSelect') {
                    console.log('[preview:selection] received selection from the editor');
                    this.setState({
                        selection: data.message
                    });
                }

                if (data.channel === 'openPage') {
                    this.setState({
                        page: cloneDeep(data.message)
                    });
                }

                if (data.channel === 'toggleShowGuides') {
                    console.warn('[preview] to be implemented');
                }

                if (data.channel === 'toggleShowHidden') {
                    console.warn('[preview] to be implemented');
                }
            });
        } else {
            console.warn('[preview:messages] failed to loop through data because it wasnt an array', e);
        }
    }

    /**
     * Callback function for window click events
     */
    listenForClick = (e) => {

        this.postToEditor('onClick', { 
            y: e.y, 
            x: e.x 
        });
    }

    /**
     * Triggered whenever any element is right clicked
     */
    contextMenuOpened( { nativeEvent: e } ) {
        if (e.shiftKey) {
            return
        }
        
        e.preventDefault();

        this.getSelectionFromElement(e.target)
            .then(selection => {
                this.setState({ selection });
                
                this.postToEditor('contextMenu', {
                    top: e.y,
                    left: e.x 
                });

            }).catch(err => console.warn('[preview:selection] no selection created'));
    }

    /**
     * Triggered whenever any element is clicked
     * @param {MouseEvent} e 
     */
    elementClicked(e) {
        console.log('[preview:click] element clicked', e.nativeEvent );
        
        this.getSelectionFromElement(e.nativeEvent.target).then( selection => {
            this.setState({ selection });
            console.log('[preview:selection] selection is ', selection);
        }).catch(err => console.warn('[preview:selection] no selection created'));
    }

    componentWillUnmount() {
        window.removeEventListener('resize', this.handleResize );
        window.removeEventListener('message', this.listenForMessages );
        window.removeEventListener('keyup', this.listenForKeyUp );
        window.removeEventListener('click', this.listenForClick );
    }

    /**
     * Posts a message to the editor
     * @param {string} channel 
     * @param {any} message 
     */
    postToEditor( channel, message ) {
        let TARGET_ORIGIN = this.props.targetOrigin;
        console.log('[preview:posting] target origin is ', this.props.targetOrigin);

        window.top.postMessage([ {
            channel, message
        }], TARGET_ORIGIN);
    }

    /**
     * This will notify the parent about our new height everytime rendering is triggered
     */
    handleResize = () => {
        console.log('[preview:resize] resize event triggered');
        /**
         * Waiting for the DOM to get rendered
         * @see https://stackoverflow.com/a/34999925/4047409
         */
        window.requestAnimationFrame(() => {
            setTimeout(() => {
                if (this.el) {
                    // We will force a re-render when a resize event is triggered because we want to ensure the focus rectangles
                    // end up placed in the correct place
                    //
                    this.setState({
                        focusRectangleKey: Math.random()
                    });
                    
                    console.log('[preview:resize] notifying parent about resize');

                    this.postToEditor('onResize', this.el.scrollHeight)

                } else {
                    console.warn('[preview:resize] attempted to resize without an element height');
                    // This function will use recursion to call this function until we have an element
                    //
                    // this.handleResize();
                }
            });
        });
    }

    /**
     * Attempst to createa a selection object out of HTML elements
     * This will return you a clean selection object with no undefined values
     * @param target {HTMLElement} the element containing the special data attributes
     * @returns { Promise<Partial<ISelection>> }
     */
    getSelectionFromElement(target) {

        console.log('[preview:selection] creating selection from element');

        /** 
         * The container index
         * @type {number}
         **/
        const containerIndex = parseInt(target.getAttribute('data-container-index'), 10);
        /** 
         * The child index
         * @type {number}
         **/
        const childIndex = parseInt(target.getAttribute('data-child-index'), 10);
        /** 
         * The block index
         * @type {number}
         **/
        const blockIndex = parseInt(target.getAttribute('data-block-index'), 10);

        /**
         * The selection object we are going to return, we are going to ensure it doesnt contain any null keys
         * We want to ensure NaN values get filtered out so we will use isFinite
         * @see https://github.com/lodash/lodash/issues/1373
         * */
        const selection = pickBy({ containerIndex, childIndex, blockIndex }, isFinite);

        // Whenever an element is clicked, the event propgates up, if we were to select a block
        // that would propogate up and select the child and its container, making the selection of the block impossible
        // an easy fix of that would be to call .stopPropagation() on the event, but that prevents the event from propogation which will cause other issues like
        // document listeners never firing because the event was unable to propogate up to the document, so a different approach will be to simply look at the created selection object
        // which will be empty if we created a selection object out of a propogated event or anything else
        //
        if (isEmpty(selection)) {
            return Promise.reject('Failed to create a selection object out of ' + target);
        } else {
            return Promise.resolve(selection);
        }
    }

    /**
     * This method will act as a callback to the FocusRectangle component, it will be called when focus on elements is calculated
     * @param {*} focusRectangle 
     */
    focusCalculated( focusRectangle ) {
        console.log('[preview:focus] focus calculated', focusRectangle);

        if ( !isEqual(focusRectangle, this.focusRectangle) ) {
            this.focusRectangle = focusRectangle;

            this.postToEditor('onSelect', {
                focusRectangle,
                selection: this.state.selection
            });
            
        } 

    }

    render() {
        console.log('[preview:rendering] rendering ');

        /** The css classes of this page */
        const pageClasses = get(this.state.page, 'CssClasses', []);

        const styleId = get(this.state.page, 'StyleSheetBlobID');

        const hostClasses = [styles.Preview, ...pageClasses].join(' ');

        let linkElement = null;

        if ( styleId ) {
            linkElement = (<link href={`${this.props.endPointUrl}/blob/${styleId}`} rel="stylesheet" id="makeup-style" />);
        }
        
        let content = [<span>Loading...</span>];


        if ( this.state.page ) {
            content = [
            (
                this.state.page.Containers.map((container, containerIndex) => {
                    // Container start
                    return <div key={containerIndex}
                        onClick={(e) => this.elementClicked(e)}
                        onContextMenu={(e) => this.contextMenuOpened(e)}
                        data-container-index={containerIndex}
                        className={`${container.CssClasses.join(' ')} eva-container-${containerIndex}`}>

                        {container.Children.map((child, childIndex) => {
                            // Child end
                            return <div
                                key={'child-' + childIndex}
                                data-container-index={containerIndex}
                                data-child-index={childIndex}
                                onClick={(e) => this.elementClicked(e)}
                                onContextMenu={(e) => this.contextMenuOpened(e)}
                                className={`${child.CssClasses.join(' ')} eva-child-${childIndex}`}>
                                {/* Blocks start */}
                                {child.Blocks.map((block, blockIndex) => {
                                    return <VisualBlock
                                        block={block}
                                        components={this.components}
                                        context={this.props.context}
                                        selector={styles['visual-block-container']}
                                        key={'block-' + blockIndex}
                                        data-container-index={containerIndex}
                                        data-child-index={childIndex}
                                        data-block-index={blockIndex}
                                        onClick={(e) => this.elementClicked(e)}
                                        onContextMenu={(e) => this.contextMenuOpened(e)}
                                    />
                                    // Blocks end
                                })}
                                {/* Child end */}
                            </div>
                        })}
                        {/* Container end */}
                    </div>
                })
            )]

            // Add a style element
            //
            content.push(
                <style>
                    {this.state.page.CssStyles}
                </style>
            )
        }

        return (
            <div>
                <div className={hostClasses} ref={(container) => this.el = container}>
                    {linkElement}

                    {content}

                    <FocusRectangle
                        key={this.state.focusRectangleKey}
                        selection={this.state.selection}
                        container={this.el}
                        focusCalculated={(e) => this.focusCalculated(e)}
                    />
                </div>
            </div>
        )
        
    }
}

export default withStyles(styles)(Preview);


