//@ts-check
import React, { Component } from 'react';
import ReactDOM from 'react-dom';
import { isEqual, cloneDeep, each, isNil, get } from 'lodash';

/**
 * This component will render components and html blocks depending on their type
 */
class VisualBlock extends Component {
    /** 
     * Element reference, this is the host of the react component we are going to render 
     * @type {HTMLElement}
     * */
    el = null;

    componentDidMount() {
        this.renderBlock(this.props);
    }

    componentWillReceiveProps(nextProps) {

        if ( !isEqual(nextProps, this.props) ) {
            console.log('[visual-block] received states, and they were not equal');
            this.renderBlock(nextProps);
        } else {
            console.warn('[visual-block] received state, and they were just equal');
        }
    }

    /**
     * This method will render a block based on the props of this component
     */
    renderBlock(props) {
        if (props.block.Type === 'component') {
            this.renderAsComponent(props);
        }

        if (props.block.Type === 'html' || props.block.Type === 'visual') {
            this.renderAsHtml(props);
        }
    }

    /**
     * Renders the props of this component as HTML
     * @param {*} props 
     */
    renderAsHtml(props) {
        console.log('[visual-block] rendering html from', props.block);
        this.el.innerHTML = props.block.Content || props.block.Template;
    }

    /**
     * Renders the props of this component as a react component
     * @param {*} props 
     */
    renderAsComponent(props) {
        console.log('[visual-block] rendering a component from', cloneDeep(props.block));
        if (props.components[props.block.ID]) {
            /**
             * The inputs of this component
             * @type { [key: string]: any }
             */
            let inputs = {};
            each(props.block.Input, (input, key) => {
                if (!isNil(input.Value)) {
                    inputs[key] = input.Value;
                }
            });

            if ( this.props.context ) { 
                inputs['store'] = this.props.context.store;
                inputs['context'] = this.props.context;
            } else {
                console.error('[visual-block] failed to find context prop');
            }
            
            console.log('[visual-block] rendering a component with inputs', inputs);
            const el = React.createElement(props.components[props.block.ID], inputs);
            ReactDOM.render(el, this.el);
        }
        else {
            console.error(`[visual-block] failed to find component ${props.block.ID}`);
        }
    }

    render() {
        let {
            'data-container-index': containerIndex,
            'data-child-index': childIndex,
            'data-block-index': blockIndex,
            /** Selector from the parent to ensure this element ends up getting the right hashed css selector */
            selector            
        } = this.props;
        
        const blockId = this.props.block.ID;

        /** Css classes of this block */
        const BlockCssClasses = get(this.props.block, 'CssClasses', []).join(' ');

        const CssClasses = `${BlockCssClasses} ${blockId} ${selector || ''}`;

        return (
            <div
                data-container-index={this.props['data-container-index']}
                data-child-index={childIndex}
                data-block-index={blockIndex}
                className={`visual-block-container eva-visual-block-${blockIndex} ${CssClasses}`}
                ref={(container) => this.el = container}>
            </div>
        );
    }
}

export default VisualBlock;
